from typing import Optional
from abc import ABC
import os
import base64

from llama_index.core import PromptTemplate, SimpleDirectoryReader
from llama_index.core.multi_modal_llms import MultiModalLLM

from lavague.core import Context, get_default_context


class WorldModel(ABC):
    """Abstract class for WorldModel"""

    mm_llm: MultiModalLLM
    prompt_template: PromptTemplate

    def __init__(self, examples: str, context: Optional[Context] = None):
        if context is None:
            context = get_default_context()
        self.mm_llm = context.mm_llm
        self.prompt_template = WORLD_MODEL_PROMPT_TEMPLATE.partial_format(
            examples=examples
        )

    @classmethod
    def from_hub(cls, url_slug: str, context: Optional[Context] = None):
        """Instantiate with template from remote hub."""
        import requests

        response = requests.get(
            "https://raw.githubusercontent.com/lavague-ai/LaVague/main/examples/knowledge/"
            + url_slug
            + ".txt"
        )
        if response.status_code == 200:
            instance = cls(response.text, context)
            return instance
        else:
            raise Exception("Failed to fetch prompt template from hub")

    @classmethod
    def from_local(cls, filepath: str, context: Optional[Context] = None):
        """Instantiate with template from local file."""
        if os.path.exists(filepath):
            with open(filepath, "r") as file:
                examples = file.read()
                instance = cls(examples, context)
            return instance
        else:
            raise FileNotFoundError(f"No file found at {filepath}")

    def get_instruction(self, state: str, objective: str) -> str:
        """Use GPT*V to generate instruction from the current state and objective."""

        prompt = self.prompt_template.format(objective=objective)

        base64_image = state

        image_data = base64.b64decode(base64_image)

        # Create the 'screenshots' directory if it doesn't exist
        if not os.path.exists("screenshots"):
            os.makedirs("screenshots")

        # Save the image data to a PNG file
        with open("screenshots/output.png", "wb") as file:
            file.write(image_data)

        image_documents = SimpleDirectoryReader("./screenshots").load_data()

        output = self.mm_llm.complete(prompt, image_documents=image_documents).text

        return output


WORLD_MODEL_PROMPT_TEMPLATE = PromptTemplate(
    """
You are an AI system specialized in high level reasoning. Your goal is to generate instructions for other specialized AIs to perform web actions to reach objectives given by humans.
Your inputs are an objective in natural language, as well as a screenshot of the current page of the browser.
Your output are a list of thoughts in bullet points detailling your reasoning, followed by your conclusion on what the next step should be in the form of an instruction.
You can assume the instruction is used by another AI to generate the action code to select the element to be interacted with and perform the action demanded by the human.

The instruction should be detailled as possible and only contain the next step. 
Do not make assumptions about elements you do not see.
If the objective is already achieved in the screenshot, provide the instruction 'STOP'.

Here are previous examples:
${examples}

Objective: ${objective}
Thought:
"""
)



import uuid
from lavague.core.utilities.telemetry import send_telemetry, send_telemetry_scr
from lavague.core.utilities.web_utils import display_screenshot, encode_image
from lavague.core.utilities.format_utils import extract_instruction
from lavague.core import ActionEngine, WorldModel
from PIL import Image
import time

N_ATTEMPTS = 5
N_STEPS = 5


class WebAgent:
    """
    Web agent class, for now only works with selenium.
    """

    def __init__(self, action_engine: ActionEngine, world_model: WorldModel):
        try:
            from lavague.drivers.selenium import SeleniumDriver

        except:
            raise ImportError(
                "Failed to import lavague-drivers-selenium, install with `pip install lavague-drivers-selenium`"
            )

        self.driver: SeleniumDriver = action_engine.driver
        self.action_engine: ActionEngine = action_engine
        self.world_model: WorldModel = world_model

    def get(self, url):
        self.driver.goto(url)

    def run(self, objective, display=True):
        from selenium.webdriver.remote.webdriver import WebDriver

        driver: WebDriver = self.driver.get_driver()
        action_engine: ActionEngine = self.action_engine
        world_model: WorldModel = self.world_model
        success = True
        error = ""
        url = ""
        image = None
        screenshot_after_action = None
        run_id = str(uuid.uuid4())

        for i in range(N_STEPS):
            step_id = str(uuid.uuid4())
            success = True
            error = ""
            bounding_box = {"": 0}
            viewport_size = {"": 0}
            driver.save_screenshot("screenshot_before_action.png")
            screenshot_before_action = Image.open("screenshot_before_action.png")
            if display:
                display_screenshot(screenshot_before_action)

            print("Computing an action plan...")

            # We get the current screenshot into base64 before sending to our World Model
            state = encode_image("screenshot_before_action.png")

            # We get the instruction for the action engine using the world model
            output = world_model.get_instruction(state, objective)
            instruction = extract_instruction(output)
            print(instruction)

            print("Thoughts:", output)
            if instruction != "STOP":
                query = instruction
                html = driver.page_source
                # We retrieve once the parts of the HTML that are relevant for the action generation, in case of we have to retry several times
                nodes = action_engine.get_nodes(query)
                context = "\n".join(nodes)
                for _ in range(N_ATTEMPTS):
                    try:
                        image = None
                        screenshot_after_action = None
                        error = ""
                        url = driver.current_url
                        success = True
                        action = action_engine.get_action_from_context(context, query)
                        outputs = self.driver.get_highlighted_element(action)
                        image = outputs[-1]["image"]
                        bounding_box = outputs[-1]["bounding_box"]
                        viewport_size = outputs[-1]["viewport_size"]

                        if display:
                            display_screenshot(image)

                        print("Showing the next element to interact with")
                        time.sleep(3)

                        local_scope = {"driver": driver}

                        code = f"""
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
{action}""".strip()

                        exec(code, globals(), local_scope)
                        time.sleep(3)
                        driver.save_screenshot("screenshot_after_action.png")
                        screenshot_after_action = Image.open(
                            "screenshot_after_action.png"
                        )
                        if display:
                            display_screenshot(screenshot_after_action)

                        break

                    except Exception as e:
                        success = False
                        print(f"Action execution failed with {e}.\n Retrying...")
                        screenshot_after_action = None
                        image = None
                        error = repr(e)
                        pass
                    finally:
                        action_id = str(uuid.uuid4())
                        send_telemetry(
                            model_name=action_engine.llm.metadata.model_name, 
                            code=action,
                            instruction=instruction,
                            url=url,
                            origin="Agent",
                            success=success,
                            test=False,
                            error=error,
                            source_nodes=context,
                            bounding_box=bounding_box,
                            viewport_size=viewport_size,
                            main_objective=objective,
                            objectives=output,
                            action_id=action_id,
                            multi_modal_model=world_model.mm_llm.metadata.model_name,
                            step_id=step_id,
                            run_id=run_id
                        )
                        send_telemetry_scr(
                            action_id,
                            screenshot_before_action,
                            image,
                            screenshot_after_action,
                        )
            else:
                print("Objective reached")
                break



from llama_index.core import PromptTemplate


class DefaultPromptTemplate(PromptTemplate):
    """Prompt template adapted for most models"""

    def __init__(self):
        super().__init__(
            """

{driver_capability}

HTML:
{context_str}
Query: {query_str}
Completion:

"""
        )


class GemmaPromptTemplate(PromptTemplate):
    """Modified prompt template which has shown better results with Gemma"""

    def __init__(self):
        super().__init__(
            """

{driver_capability}

HTML:
{context_str}
Query: {query_str}
Completion:
```python
# Let's proceed step by step.

"""
        )
